/*
 * 네오 대포 SOURCE CODE
 * Author: TaeJune Joung
 * Enterprise: NEO3DS
 * Date: 22.01.17
 * Update: mj
 *   - 통신방법 변경
 *     (잡음이 끼는 것을 방지하고자 처리 -> 앱 통신이 깔끔해짐/엔트리 보드로 연결도 가능)
 *   
 *  [컨트롤러 소스 통합] 23.04.17
 *   - Mode 추가
 *   
 *  [데이터 길이 수정] 22.12.28
 *   - MOTOR SPEED 데이터 삭제
 *   
 *  [보드형 -> 완성형] 22.06.24
 *   - RGB LED -> LED 변경
 *   - 발사모터/각도모터 제어 추가
 *   
 *  [컨트롤러 소스 통합]
 *   - Mode 추가
 *   
 * License:
 *  본 코드의 저작권은 위와 같으며,
 *  상업적인 무단 도용시 법적인 책임을 물을 수 있음을 공지합니다.
 */
 
/* MODE 상수 */
#define CONTROL        0
#define BLOCKLY        1

/* 동작 상수 */
#define ALIVE          0
#define VIBE           1

/* 상태 상수 */
#define GET            1
#define SET            2
#define RESET          3

/* 투석기 구동 모터 */
#define PIN_LEFT_MOTOR_FW      2
#define PIN_LEFT_MOTOR_BW      4
#define PIN_RIGHT_MOTOR_FW     8
#define PIN_RIGHT_MOTOR_BW     7

/* LED 핀 */
#define PIN_LED               11

/* 버저 핀 */
#define PIN_BUZZER             3
/* 진동센서 핀 */
#define PIN_VIBE              A0

/* 장전 모터 */
#define PIN_GEAR_MOTOR_FW      9
#define PIN_GEAR_MOTOR_BW     10

/* 각도 모터 */
#define PIN_ANGLE_MOTOR_FW     6
#define PIN_ANGLE_MOTOR_BW     5

/* 
 *전역변수 선언
 */
int mode             = 0;
int life             = 3;
int shootMotorState  = 0;

/* 장전기어 모터 속도 */
int GEAR_MAX_SPEED = 200;
int GEAR_MIN_SPEED = 0;
int GEAR_CATCH_SPEED = 80;

/* 각도기어 모터 속도 */
int ANGLE_MAX_SPEED = 220;
int ANGLE_MIN_SPEED = 0;

union {
  byte byteVal[4];
  float floatVal;
  long longVal;
} val;

union {
  byte byteVal[2];
  short shortVal;
} valShort;

const int START_VAL = 255;
const int DATA_LEN = 14;
const int END_VALUE = 10;

int beforeBuzNote = 0;
int rotation = 1000;

unsigned char buf[DATA_LEN];
unsigned char prevc=0;
int index = 0;
boolean isStart = false;

int vibeState = 0;
boolean isReload = false;

void initPorts() {
  for (int i=2; i < 14; i++) {
    pinMode(i, OUTPUT);
    digitalWrite(i, LOW);
  }
  // 각도 조절 모터 - Cannon
  pinMode(PIN_ANGLE_MOTOR_FW, OUTPUT);
  pinMode(PIN_ANGLE_MOTOR_BW, OUTPUT);
  digitalWrite(PIN_ANGLE_MOTOR_FW, LOW);
  digitalWrite(PIN_ANGLE_MOTOR_BW, LOW);
}

void updateDigitalPort(unsigned char c) {
  if (c == DATA_LEN && isStart == false) {
    if (prevc == START_VAL) {
      isStart = true;
      index = 1;
      writeBuffer(index-1, prevc);
      writeBuffer(index, c);
    }
  } else {
    prevc = c;
    if (isStart) {
      writeBuffer(index, c);
    }
  }

  index++;
  
  if (index > DATA_LEN) {
    isStart = false;
    index = 0;
  }

  if (isStart && index == DATA_LEN && index > 3) {
    isStart = false;
    parseData();
    index = 0;
  }
}

boolean isCheckOk () {
  if (buf[DATA_LEN-1] == END_VALUE) {
    int checkSum = buf[DATA_LEN-2];
    int sumValue = 0;
    for (int i=2; i < DATA_LEN-2; i++) {
      sumValue += buf[i];
    }
    sumValue = sumValue & 0xff;
    if (checkSum == sumValue) {
      return true;
    }
  }
  return false;
}

unsigned char readBuffer(int index) {
  return buf[index];
}

void buzCtrl(int buzOctave, int buzNote) {
  int note = (buzOctave << 8) + buzNote;
  if (beforeBuzNote != note) {
    beforeBuzNote = note;
    if (note > 30) {
      tone(PIN_BUZZER, note);
    } else {
      noTone(PIN_BUZZER);
    }
  }
}

void motorCtrl(int motorState) {
  switch(motorState) {
    case 0: //Stop
      digitalWrite(PIN_RIGHT_MOTOR_FW, LOW);
      digitalWrite(PIN_RIGHT_MOTOR_BW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_FW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_BW, LOW);
      break;
    case 1: // Forward
      digitalWrite(PIN_RIGHT_MOTOR_FW, HIGH);
      digitalWrite(PIN_RIGHT_MOTOR_BW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_FW, HIGH);
      digitalWrite(PIN_LEFT_MOTOR_BW, LOW);
      break;
    case 2: // Turn Left
      digitalWrite(PIN_RIGHT_MOTOR_FW, HIGH);
      digitalWrite(PIN_RIGHT_MOTOR_BW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_FW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_BW, HIGH);
      break;
    case 3: // Turn Right
      digitalWrite(PIN_RIGHT_MOTOR_FW, LOW);
      digitalWrite(PIN_RIGHT_MOTOR_BW, HIGH);
      digitalWrite(PIN_LEFT_MOTOR_FW, HIGH);
      digitalWrite(PIN_LEFT_MOTOR_BW, LOW);
      break;
    case 4: // Backward
      digitalWrite(PIN_RIGHT_MOTOR_FW, LOW);
      digitalWrite(PIN_RIGHT_MOTOR_BW, HIGH);
      digitalWrite(PIN_LEFT_MOTOR_FW, LOW);
      digitalWrite(PIN_LEFT_MOTOR_BW, HIGH);
      break;
  }
}

void ledCtrl(int ledPin, int ledValue) {
  if (ledValue == 255) {
    digitalWrite(ledPin, HIGH);
  } else if (ledValue == 0) {
    digitalWrite(ledPin, LOW);
  } else {
    analogWrite(ledPin, ledValue);
  }
}

void controlShootMotorCtrl() {
  if (shootMotorState == 2) {
    shootMotorCtrl(shootMotorState);
    delay(3000);
  }
  while (shootMotorState == 2 && life > 0) {
    int vibe_value =  analogRead(PIN_VIBE);
    if (vibe_value > 600) {
      writeValue(0xff);
      writeValue(0x7);
      writeValue(VIBE);
      writeValue(1); // 감지=1 / 감지x=0
      writeValue(--life);
      int checkSum = 0;
      checkSum = (VIBE + 1 + life) & 0xff;
      writeValue(checkSum);
      writeEnd();

      for(int i=0; i < 2; i++) {
        tone(PIN_BUZZER, 194, 100);
        delay(150);
        tone(PIN_BUZZER, 212, 100);
        delay(150);
      }
      delay(2000);

      if (life == 0) {
        digitalWrite(PIN_LED, LOW);
    
        for(int i=0; i < 10; i++) {
          tone(PIN_BUZZER, 33, 300);
          delay(350);
        }
        life -= 1;
      } else if (life == 3) {
        analogWrite(PIN_LED, 255);
      } else if (life == 2) {
        analogWrite(PIN_LED, 168);
      } else if (life == 1) {
        analogWrite(PIN_LED, 84);
      }
    }
    bluetooth_read();
  }
  if (shootMotorState != 2) {
    shootMotorCtrl(shootMotorState);
  }
}

void shootMotorCtrl(int shootMotorState) {
  if (shootMotorState == 0) { // 정지
    analogWrite(PIN_GEAR_MOTOR_FW, GEAR_MIN_SPEED);
    analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MIN_SPEED);
  } else if (shootMotorState == 1) { // 정회전
    isReload = true;
    analogWrite(PIN_GEAR_MOTOR_FW, GEAR_MAX_SPEED);
    analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MIN_SPEED);
  } else if (shootMotorState == 2) { // 역회전
    if (isReload) {
      analogWrite(PIN_GEAR_MOTOR_FW, GEAR_MIN_SPEED);
      analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MAX_SPEED);
      delay(80);
      isReload = false;
    }
    analogWrite(PIN_GEAR_MOTOR_FW, GEAR_MIN_SPEED);
    analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MIN_SPEED);
  } else if (shootMotorState == 3) { // 잡아주기
    analogWrite(PIN_GEAR_MOTOR_FW, GEAR_CATCH_SPEED);
    analogWrite(PIN_GEAR_MOTOR_BW, GEAR_MIN_SPEED);
  }
}

void angleMotorCtrl(int angleMotorState) {
  if (angleMotorState == 1) { // 정회전
    analogWrite(PIN_ANGLE_MOTOR_FW, ANGLE_MAX_SPEED);
    analogWrite(PIN_ANGLE_MOTOR_BW, ANGLE_MIN_SPEED);
  } else if (angleMotorState == 2) { // 역회전
    analogWrite(PIN_ANGLE_MOTOR_FW, ANGLE_MIN_SPEED);
    analogWrite(PIN_ANGLE_MOTOR_BW, ANGLE_MAX_SPEED);
  } else { // 정지
    digitalWrite(PIN_ANGLE_MOTOR_FW, ANGLE_MIN_SPEED);
    digitalWrite(PIN_ANGLE_MOTOR_BW, ANGLE_MIN_SPEED);
  }
}

void parseData() {
  if (isCheckOk()) {
    mode = readBuffer(2);
    life = readBuffer(3);
    int buzOctave = readBuffer(4);
    int buzNote = readBuffer(5);
    int motorState = readBuffer(6);
    int ledValue = readBuffer(7);
    shootMotorState = readBuffer(8);
    int d9Value = readBuffer(9);
    int d10Value = readBuffer(10);
    int angleMotorState = readBuffer(11);
    
    buzCtrl(buzOctave, buzNote);
    motorCtrl(motorState);
    ledCtrl(PIN_LED, ledValue);

    if (mode == 0) {
      controlShootMotorCtrl();
    } else {
      if (shootMotorState == 4) {
        ledCtrl(PIN_GEAR_MOTOR_FW, d9Value);
        ledCtrl(PIN_GEAR_MOTOR_BW, d10Value);
      } else {
        shootMotorCtrl(shootMotorState);
      }
    }
    angleMotorCtrl(angleMotorState);
  }
}

void getVibeValue() {
  writeValue(0xff);
  writeValue(0x7);
  writeValue(VIBE);
  int vibeValue = analogRead(PIN_VIBE);
  if (vibeValue > 600) {
    vibeState = 1;
  } else {
    vibeState = 0;
  }
  writeValue(vibeState);
  writeValue(life);
  int checkSum = 0;
  checkSum = (VIBE + vibeState + life) & 0xff;
  writeValue(checkSum);
  writeEnd();
}

void sendSensorValues() {
  getVibeValue();
}

void writeBuffer(int index, unsigned char c) {
  buf[index] = c;
}

void writeValue(unsigned char c) {
  Serial.write(c);
}

void writeEnd() {
  Serial.write(0xd);
  Serial.write(0xa);
}

void sendFloat(float value) {
  val.floatVal = value;
  writeValue(val.byteVal[0]);
  writeValue(val.byteVal[1]);
  writeValue(val.byteVal[2]);
  writeValue(val.byteVal[3]);
}

void bluetooth_read() {
  while (Serial.available()) {
    unsigned char c = Serial.read();
    Serial.println(c);
    updateDigitalPort(c);
  }
}

void setup() {
  initPorts();
  Serial.begin(115200);

  /* 작동 소리 */
  tone(PIN_BUZZER, 912, 100);
  delay(150);
  tone(PIN_BUZZER, 912, 100);
  delay(150);

  /* [TODO]작동 소리 2안 */
//  tone(PIN_BUZZER, 723, 100);
//  delay(150);
//  tone(PIN_BUZZER, 836, 100);
//  delay(150);
//  tone(PIN_BUZZER, 912, 100);
//  delay(150);
}

void loop() {
  bluetooth_read();

  if (mode == 1) {
    if(rotation > 1000) {
      rotation = 0;
      sendSensorValues();
    }
    rotation++;
  }
}

void callOK() {
  writeValue(0xff);
  writeValue(0xb);
  writeEnd();
}
